"""
Assign custom values to quantities
"""


#[
from __future__ import annotations

from typing import (TYPE_CHECKING, Iterable )

from ..wrongdoings import (IrisPieCritical, )
from ..conveniences import iterators as _iterators
from ..databoxes.main import (Databox, )

if TYPE_CHECKING:
    from typing import (Any, )
    from collections.abc import (Iterator, )
    from numbers import (Real, )
#]


# TODO: Use in Sequentials


class Inlay:
    """
    """
    #[

    def assign(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
        Assign custom values to quantities
        """
        assigned_keys, nonexistent_keys = self._assign(*args, **kwargs, )

    def assign_strict(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
        Assign custom values to quantities
        """
        assigned_keys, nonexistent_keys = self._assign(*args, **kwargs, )
        if nonexistent_keys:
            message = ("Cannot assign these names (nonexistent in the model object): ", ) + nonexistent_keys
            raise IrisPieCritical(message, )

    def _assign(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
        Assign parameters from dicts or from keyword arguments
        """
        dict_to_assign = {}
        dict_to_assign.update(*args, )
        dict_to_assign.update(kwargs, )
        if not dict_to_assign:
            return
        #
        name_to_qid = self.create_name_to_qid()
        qid_to_name = self.create_qid_to_name()
        qid_to_custom_values, assigned_keys, nonexistent_keys \
            = _rekey_dict(dict_to_assign, name_to_qid, )
        qid_to_custom_values_iter = Databox.iter_variants(
            qid_to_custom_values,
            item_iterator=_prepare_custom_value_iter,
        )
        for variant, values in zip(self._variants, qid_to_custom_values_iter, ):
            variant.update_values_from_dict(values, )
            self._enforce_assignment_rules(variant, )
        #
        return assigned_keys, nonexistent_keys
    #]


def _rekey_dict(
    dict_to_rekey: dict,
    old_key_to_new_key: dict,
) -> tuple[dict[int, Any], tuple[str, ...], tuple[str, ...]]:
    """
    """
    #[
    new_dict = {}
    assigned_keys = set()
    nonexistent_keys = set()
    for key, value in dict_to_rekey.items():
        try:
            new_key = old_key_to_new_key[key]
            new_dict[new_key] = value
            assigned_keys.add(key)
        except KeyError:
            nonexistent_keys.add(key)
    return new_dict, tuple(assigned_keys), tuple(nonexistent_keys)
    #]


_UNCHANGED_VALUE = (..., ..., )


def _prepare_custom_value_iter(
    value: Any,
    /,
) -> Iterator[tuple[Real | EllipsisType, Real | EllipsisType]]:
    """
    Resolve custom values: value = [a, b, c] means variants, a = (1, 2) means level and change, ... means keep unchanged
    """
    #[
    is_iterable = (
        isinstance(value, Iterable, )
        and not isinstance(value, str, )
        and not isinstance(value, bytes, )
        and not isinstance(value, tuple, )
    )
    value = value if is_iterable else [value, ]
    return _iterators.exhaust_then_last(value, _UNCHANGED_VALUE, )
    #]




"""
Implement SimulationPlannableProtocol
"""


#[
from __future__ import annotations
#]


class _SimulationPlannable:
    """
    """
    #[

    def __init__(
        self,
        sequential,
        **kwargs,
    ) -> None:
        """
        """
        self.can_be_exogenized \
            = tuple(set(
                i.lhs_name for i in sequential._invariant.explanatories
                if not i.is_identity
            ))
        self.can_be_endogenized \
            = tuple(set(
                i.residual_name for i in sequential._invariant.explanatories
                if not i.is_identity
            ))


class Inlay:
    """
    """
    #[

    def get_simulation_plannable(self, **kwargs, ) -> _SimulationPlannable:
        """
        """
        return _SimulationPlannable(self, **kwargs, )

    #]




"""
"""

#[
from __future__ import annotations

import dill as _di
import json as _js
#]


__all__ = (
    "save",
    "load",
    "save_json",
    "load_json",
)


def save(filename, obj, ):
    with open(filename, "wb", ) as fid:
        _di.dump(obj, fid, )


def load(filename, ):
    with open(filename, "rb", ) as fid:
        return _di.load(fid, )


def save_json(filename: str, something, *, indent: int = 4, **kwargs, ):
    """
    """
    with open(filename, "wt", ) as fid:
        _js.dump(something, fid, indent=4, **kwargs, )


def load_json(filename: str, **kwargs, ):
    """
    """
    with open(filename, "rt", ) as fid:
        return _js.load(fid, )





"""
Time series mixin for moving sum, average, product
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Self, Callable, )
import numpy as _np
import functools as _ft

from .. import pages as _pages
from .. import dates as _dates
from . import _functionalize
#]


__all__ = ()


class Inlay:
    """
    Time series mixin for moving sum, average, product
    """
    #[

    @_pages.reference(
        category=None,
        call_name="Moving window calculations",
        call_name_is_code=False,
        priority=20,
    )
    def moving_window(
        self,
        func: Callable,
        /,
        window: int | None = None,
    ) -> Self:
        r"""
················································································


Overview of moving window functions:

| Function | Description
|----------|-------------
| `mov_sum` | Moving sum, $y_t = \sum_{i=0}^{k-1} x_{t-i}$
| `mov_avg` | Moving average, $y_t = \frac{1}{k} \sum_{i=0}^{k-1} x_{t-i}$
| `mov_mean` | Same as `mov_avg`
| `mov_prod` | Moving product, $y_t = \prod_{i=0}^{k-1} x_{t-i}$

where

* $k$ is the window length determined by the `window` input argument
(note that $k$ above is a positive integer while `window` needs to be
entered as a negative integer).



### Function for creating new Series objects ###

```
new = irispie.mov_sum(self, window=None, )
new = irispie.mov_avg(self, window=None, )
new = irispie.mov_mean(self, window=None, )
new = irispie.mov_prod(self, window=None, )
```


### Methods for changing existing Series objects in-place ###


```
self.mov_sum(window=None, )
self.mov_avg(window=None, )
self.mov_mean(window=None, )
self.mov_prod(window=None, )
```


### Input arguments ###


???+ input "self"
    Time series on whose data a moving function is calculated (see the
    overview table above).

???+ input "window"
    A negative interger determining the number of observations to include
    in the moving window, counting from the current time period backwards
    (the minus sign is a convention to indicate that the window goes
    backwards in time). If `window=None` (or not specified), the default
    window is derived from the time series frequency:

    | Date frequency | Default window |
    |----------------|---------------:|
    | `YEARLY`       |             –1 |
    | `QUARTERLY`    |             –4 |
    | `MONTHLY`      |            –12 |
    | `WEEKLY`       |            –52 |
    | `DAILY`        |           –365 |
    | Otherwise      |             –4 |

    Note that the default windows for `DAILY` and `WEEKLY` frequencies are
    fixed numbers and do not depend on the actual number of days in a
    calendar year or the actual number of weeks in a calendar year.


### Returns ###


???+ returns "new"
    New time series with data calculated as a moving window function of the
    original data.

???+ returns "self"
    Time series with data replaced by the moving window function of the
    original data.

················································································
        """
        window = (
            window
            if window is not None
            else self._get_default_moving_window()
        )
        window_length = -window
        data = _np.pad(
            self.data,
            pad_width=((window_length-1, 0), (0, 0)),
            mode="constant",
            constant_values=_np.nan,
        )
        data_windows =_np.lib.stride_tricks.sliding_window_view(
            data,
            window_shape=window_length,
            axis=0,
        )
        new_data = func(data_windows, axis=2, )
        self._replace_data(new_data, )

    @_pages.reference(category="moving", )
    def mov_sum(self, window: int | None = None, ) -> Self:
        r"""
................................................................................

==Moving sum==


See documentation of [moving window calculations](#moving-window-calculations) or
`help(irispie.Series.moving_window)`.

................................................................................
        """
        return self.moving_window(_np.sum, window=window, )

    @_pages.reference(category="moving", )
    def mov_avg(self, window: int | None = None, ) -> Self:
        r"""
................................................................................

==Moving average==


See documentation of [moving window calculations](#moving-window-calculations) or
`help(irispie.Series.moving_window)`.

................................................................................
        """
        return self.moving_window(_np.mean, window=window, )

    @_pages.reference(category="moving", )
    def mov_mean(self, window: int | None = None, ) -> Self:
        r"""
................................................................................

==Moving average==


See documentation of [moving window calculations](#moving-window-calculations) or
`help(irispie.Series.moving_window)`.

................................................................................
        """
        return self.moving_window(_np.mean, window=window, )

    @_pages.reference(category="moving", )
    def mov_prod(self, window: int | None = None, ) -> Self:
        r"""
................................................................................

==Moving product==


See documentation of [moving window calculations](#moving-window-calculations) or
`help(irispie.Series.moving_window)`.

................................................................................
        """
        return self.moving_window(_np.prod, window=window, )

    def _get_default_moving_window(self, /, ) -> int:
        r"""
        Derive default moving window from time series frequency
        """
        return (
            -self.frequency.value
            if self.frequency is not None and self.frequency.value > 0
            else -4
        )

    #]


for n in ("mov_sum", "mov_avg", "mov_mean", "mov_prod", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )




"""
"""


#[
from __future__ import annotations
from numbers import (Number, )
import numpy as _np
import re as _re

from ..conveniences import views as _views
#]


_REPR_INDENT = "⏐ "
_REPR_SEPARATOR = ": "
_REPR_MAX_LEN = 50
_REPR_CONT = "…"


class Inlay(_views.Mixin, ):
    """
    """
    #[

    def _get_first_line_view(self, /, ):
        """
        """
        return f"{self.__class__.__name__} with {self.num_items:g} item(s)"

    def _get_header_separator(self, /, ):
        """"
        """
        return _REPR_INDENT.rstrip()

    def _get_content_view(self, /, ):
        """
        """
        names = self.get_names()
        if names:
            max_len = self._get_max_name_length_()
            content = tuple(_REPR_INDENT + str(k).rjust(max_len) + _REPR_SEPARATOR + _databox_repr(self[k]) for k in names)
        else:
            content = ()
        return content

    def _get_max_name_length_(self, /, ):
        """
        """
        return max(len(str(k)) for k in self.get_names())

    def _get_short_row_(self):
        """
        """
        max_len = self._get_max_name_length_()
        return _REPR_INDENT + _views._VERTICAL_ELLIPSIS.rjust(max_len) + " "*len(_REPR_SEPARATOR) + _VERTICAL_ELLIPSIS

    def _get_footer_view_(self, /, ):
        return (_REPR_INDENT.rstrip(), )
    #]


def _databox_repr(x, /, ) -> str:
    """
    String representing one item in a databox
    """
    #[

    if hasattr(x, "_databox_repr"):
        s = f"<{x._databox_repr()}>"
    elif x is None:
        s = "None"
    elif x is ...:
        s = "..."
    elif isinstance(x, Number):
        s = str(x)
    elif isinstance(x, str):
        s = f'"{x}"'
    elif isinstance(x, _np.ndarray) or isinstance(x, list) or isinstance(x, tuple):
        s = _re.sub(r"\n+ +", " ", repr(x))
    elif hasattr(x, "_get_first_line_view"):
        s = f"<{x._get_first_line_view()}>"
    else:
        s = repr(type(x))
    return s if len(s)<_REPR_MAX_LEN else s[0:_REPR_MAX_LEN] + _REPR_CONT

    #]





import pytest
import calendar as ca
import random as rn
import numpy as np
import irispie as ir

rn.seed(0)


daily_span = ir.dd(2020,1,1) >> ir.dd(2021,12,31)

parameter_names = "method, result_func"

parameters = [
    ("sum", lambda num_days: (1 + num_days) * num_days / 2),
    ("mean", lambda num_days: (1 + num_days) / 2),
    (None, lambda num_days: (1 + num_days) / 2),
    ("first", lambda num_days: 1),
    ("last", lambda num_days: num_days),
]

# Daily to yearly

montly_span = ir.mm(2020,1) >> ir.mm(2021,12)
monthly_period_value = ((t, t.day) for t in daily_span)
periods, values = zip(*monthly_period_value)
DAILY_TO_MONTHLY_SERIES = ir.Series(periods=periods, values=values, )

@pytest.mark.parametrize(parameter_names, parameters, )
def test_daily_to_monthly(method, result_func, d=DAILY_TO_MONTHLY_SERIES, ):
    m = ir.aggregate(d, ir.MONTHLY, method=method, )
    for t in montly_span:
        _, num_days = ca.monthrange(t.year, t.period)
        assert m[t] == result_func(num_days, )


# Daily to yearly


def day_in_year(t):
    return t.serial - ir.daily_serial_from_ymd(t.year, 1, 1) + 1
yearly_span = ir.yy(2020) >> ir.yy(2021)
yearly_period_value = ((t, day_in_year(t)) for t in daily_span)
periods, values = zip(*yearly_period_value)
DAILY_TO_YEARLY_SERIES = ir.Series(periods=periods, values=values, )


@pytest.mark.parametrize(parameter_names, parameters, )
def test_daily_to_yearly(method, result_func, d=DAILY_TO_YEARLY_SERIES, ):
    m = ir.aggregate(d, ir.YEARLY, method=method, )
    for t in yearly_span:
        num_days = (ca.isleap(t.year) and 366) or 365
        assert m[t] == result_func(num_days, )


daily_span = ir.dd(2020,1,1) >> ir.dd(2021,12,31)
montly_span = ir.mm(2020,1) >> ir.mm(2021,12)
monthly_period_value = ((t, t.day if t.day != 15 else None) for t in daily_span)
periods, values = zip(*monthly_period_value)
DAILY_SERIES_MISSING = ir.Series(periods=periods, values=values, )


parameter_names = "method, result_func"

parameters = [
    ("sum", lambda num_days: sum(range(1, num_days + 1)) - 15),
    ("mean", lambda num_days: (sum(range(1, num_days + 1)) - 15) / (num_days - 1)),
    (None, lambda num_days: (sum(range(1, num_days + 1)) - 15) / (num_days - 1)),
    ("first", lambda num_days: 1),
    ("last", lambda num_days: num_days),
]

@pytest.mark.parametrize(parameter_names, parameters, )
def test_daily_to_monthly_missing(method, result_func, d=DAILY_SERIES_MISSING, ):
    m = ir.aggregate(d, ir.MONTHLY, method=method, remove_missing=True, )
    for t in montly_span:
        _, num_days = ca.monthrange(t.year, t.period)
        assert m[t] == result_func(num_days, )

